package net.xinhuamm.xycloud.xinhuamm_xycloud;

import android.app.Activity;
import android.app.Dialog;
import android.content.Context;
import android.content.DialogInterface;
import android.content.pm.ActivityInfo;
import android.content.res.Resources;
import android.graphics.Color;
import android.graphics.drawable.ColorDrawable;
import android.graphics.drawable.GradientDrawable;
import android.graphics.drawable.InsetDrawable;
import android.os.Bundle;
import android.view.LayoutInflater;
import android.view.View;
import android.view.ViewGroup;
import android.view.ViewStub;
import android.view.Window;
import android.view.WindowManager;

import androidx.annotation.ColorRes;
import androidx.annotation.FloatRange;
import androidx.annotation.IdRes;
import androidx.annotation.LayoutRes;
import androidx.annotation.NonNull;
import androidx.annotation.Nullable;
import androidx.core.content.ContextCompat;
import androidx.fragment.app.DialogFragment;
import androidx.fragment.app.FragmentManager;

/**
 * created by cmj on 2019-07-16
 * for:弹出窗基类
 */
public abstract class BaseDialog extends DialogFragment {

    protected Context context;
    private boolean isInflate;
    protected View rootView;
    protected boolean isLandscape;
    protected int dialogWidth;

    /**
     * 消失监听
     */
    private DialogInterface.OnDismissListener onDismissListener;
    /**
     * 显示监听
     */
    private DialogInterface.OnShowListener onShowListener;

    public BaseDialog() {
        dialogWidth = (int) dpToPx(250);//默认宽度-防止onCreate中赋值失败
    }

    @Override
    public void onAttach(@NonNull Context context) {
        super.onAttach(context);
        this.context = context;
        isLandscape = this.context instanceof Activity && ((Activity) this.context).getRequestedOrientation()
                == ActivityInfo.SCREEN_ORIENTATION_LANDSCAPE;
    }

    @Override
    public void onCreate(@Nullable Bundle savedInstanceState) {
        super.onCreate(savedInstanceState);
        setDialogTheme();
    }

    protected void setDialogTheme() {
        setStyle(DialogFragment.STYLE_NO_TITLE, getDialogTheme());
    }

    protected int getDialogTheme() {
        return R.style.Theme_Dialog_Base;
    }

    @Override
    public void onStart() {
        super.onStart();
        Dialog dialog = getDialog();
        if (null != dialog) {
            Window window = dialog.getWindow();
            if (null != window) {
                onStartWindowConfig(window);
            }
        }
    }

    /**
     * 设置窗口相关属性
     *
     * @param window window
     */
    protected void onStartWindowConfig(@NonNull Window window) {
        window.setBackgroundDrawable(new ColorDrawable(Color.TRANSPARENT));//设置dialog默认背景的样式,系统自带背景不可编辑
        WindowManager.LayoutParams params = window.getAttributes();
        params.dimAmount = getBackgroundAmount();
        params.height = WindowManager.LayoutParams.WRAP_CONTENT;
        params.width = dialogWidth;
        window.setAttributes(params);
    }

    @Override
    public View onCreateView(@NonNull LayoutInflater inflater, ViewGroup container,
                             Bundle savedInstanceState) {
        if (null == rootView) {
            rootView = inflater.inflate(getRootLayoutId(), container);
            InsetDrawable insetDrawable = new InsetDrawable(getGradientDrawable(),
                    getContentLeftMargin(), getContentTopMargin(),
                    getContentRightMargin(), getContentBottomMargin());
            rootView.setBackground(insetDrawable);
            ViewStub stub = rootView.findViewById(R.id.dialog_stub);
            stub.setOnInflateListener((stub1, inflated) -> isInflate = true);
            stub.setLayoutResource(getContentLayoutId());
            if (!isInflate) {
                stub.inflate();
            }
            initBundle(getArguments());
            initWidget(savedInstanceState);
            Dialog dialog = getDialog();
            if (null != dialog) {
                initDialog(dialog);
                dialog.setOnShowListener(onShowListener);
            }
        } else {
            if (rootView.getParent() != null) {
                ((ViewGroup) rootView.getParent()).removeView(rootView);
            }
        }
        return rootView;
    }

    protected int getRootLayoutId() {
        return R.layout.xinhuamm_xycloud_layout_dialog_base_center;
    }

    protected final <T extends View> T findViewById(@IdRes int id) {
        return rootView.findViewById(id);
    }

    protected void initBundle(@Nullable Bundle bundle) {

    }

    protected void initWidget(Bundle savedInstanceState) {

    }

    protected void initDialog(@NonNull Dialog dialog) {

    }

    /**
     * 透明度
     */
    @FloatRange(from = 0.0, to = 1.0)
    protected float getBackgroundAmount() {
        return 0.4f;
    }

    /**
     * 内容布局
     *
     * @return 内容布局
     */
    @LayoutRes
    protected abstract int getContentLayoutId();

    /**
     * 内容背景颜色
     *
     * @return 内容背景颜色
     */
    @ColorRes
    protected abstract int getBackgroundColorRes();

    /**
     * 内容背景Drawable
     */
    protected GradientDrawable getGradientDrawable() {
        GradientDrawable shape = new GradientDrawable();
        float corner10 = dpToPx(10);
        shape.setCornerRadii(new float[]{corner10, corner10, corner10, corner10,
                corner10, corner10, corner10, corner10});
        shape.setColor(ContextCompat.getColor(context, getBackgroundColorRes()));
        return shape;
    }

    /**
     * 根据手机的分辨率从 dp 的单位 转成为 px(像素)
     *
     * @param dpValue 虚拟像素
     * @return 像素
     */
    protected float dpToPx(float dpValue) {
        return (0.5f + dpValue * Resources.getSystem().getDisplayMetrics().density);
    }

    /**
     * 根内容左边距
     *
     * @return 根内容左边距
     */
    protected int getContentLeftMargin() {
        return 0;
    }

    /**
     * 根内容上边距
     *
     * @return 根内容上边距
     */
    protected int getContentTopMargin() {
        return 0;
    }

    /**
     * 根内容右边距
     *
     * @return 根内容右边距
     */
    protected int getContentRightMargin() {
        return 0;
    }

    /**
     * 根内容下边距
     *
     * @return 根内容下边距
     */
    protected int getContentBottomMargin() {
        return 0;
    }

    /**
     * 设置消失监听器
     */
    public void setOnDismissListener(DialogInterface.OnDismissListener mOnDismissListener) {
        this.onDismissListener = mOnDismissListener;
    }

    /**
     * 设置显示监听器
     */
    public void setOnShowListener(DialogInterface.OnShowListener mOnShowListener) {
        this.onShowListener = mOnShowListener;
    }

    /**
     * 隐藏
     *
     * @param dialog d
     */
    @Override
    public void onDismiss(@NonNull DialogInterface dialog) {
        super.onDismiss(dialog);
        if (null != onDismissListener) {
            onDismissListener.onDismiss(dialog);
        }
    }

    /**
     * 显示
     *
     * @param fragmentManager fm
     */
    public void show(FragmentManager fragmentManager) {
        show(fragmentManager, this.getClass().getSimpleName());
    }

    /**
     * @return 是否显示中
     */
    public boolean isShowing() {
        return getDialog() != null && getDialog().isShowing();
    }

}
